# encoding: utf-8
import logging
from pdefc.lang.common import Located


class AbstractImport(Located):
    '''AbstractImport is a base class for module imports.'''
    def __init__(self, location=None):
        self.location = location
        self.alias_module_pairs = []

    def link(self, package):
        '''Link this import and return a list of errors.'''
        return []


class AbsoluteImport(AbstractImport):
    '''AbsoluteImport references a single module by its absolute name.'''
    def __init__(self, name, location=None):
        super(AbsoluteImport, self).__init__(location=location)
        self.name = name

    def __str__(self):
        return 'import ' + self.name

    def __repr__(self):
        return '<%s %s at %s>' % (self.__class__.__name__, self.name, hex(id(self)))

    def link(self, package):
        logging.debug('Linking %s', self)

        module = package.lookup_module(self.name)
        if not module:
            return [self._error('Module not found "%s"', self.name)]

        self.alias_module_pairs = [(self.name, module)]
        return []


class RelativeImport(AbstractImport):
    '''RelativeImport references modules with a prefix and multiple relative names,
    i.e, from my_package import module0, module1.'''
    def __init__(self, prefix, relative_names, location=None):
        super(RelativeImport, self).__init__(location=location)

        self.prefix = prefix
        self.relative_names = tuple(relative_names)

    def __str__(self):
        return '"from ' + self.prefix + ' import ' + ', '.join(self.relative_names) + '"'

    def __repr__(self):
        return '<%s %s at %s>' % (self.__class__.__name__, self, hex(id(self)))

    def link(self, package):
        logging.debug('Linking \'%s\'', self)

        errors = []
        for rname in self.relative_names:
            fullname = self.prefix + '.' + rname

            module = package.lookup_module(fullname)
            if module:
                self.alias_module_pairs.append((rname, module))
            else:
                errors.append(self._error('Module not found "%s"', fullname))

        return errors



# encoding: utf-8
from __future__ import unicode_literals
import functools
import logging
import re
import ply.lex as lex
import ply.yacc as yacc

import pdefc.lang
from pdefc.reserved import RESERVED


def create_parser():
    '''Create a new parser, the parser is reusable but not thread-safe.'''
    return Parser()


class Parser(object):
    '''Pdef parser. It is reusable but not thread-safe.'''

    def __init__(self):
        self.grammar = _Grammar(self._name, self._error)

        # Some docs on options:
        # - optimize=False and write_tables=False force to generate tabmodule each time
        #   parser is created. It is required for production.
        # - module=self.grammar sets the grammar for a lexer and a parser.
        # - start='file' sets the start grammar rule.

        self.lexer = lex.lex(module=self.grammar, optimize=False, debug=False, reflags=re.UNICODE)
        self.parser = yacc.yacc(module=self.grammar, optimize=False, write_tables=False,
                                start='module', debug=False)

        # These are cleaned on each parse invocation.
        self._errors = []
        self._module_name = None
        self._path = None

    def parse(self, source, relative_name, path=None):
        '''Parse a module from a source string, return the module and a list of errors.'''
        name = relative_name
        logging.info('Parsing %s', path or name)

        # Clear the variables.
        self._errors = []
        self._module_name = name
        self._path = path

        try:
            lexer = self.lexer.clone()
            module = self.parser.parse(source, tracking=True, lexer=lexer)
            errors = list(self._errors)

            if module:
                module.relative_name = name
                module.path = path

            if errors:
                self._log_errors(errors, path or name)
                return None, errors

            return module, errors

        finally:
            self._errors = None
            self._path = None

    def _name(self):
        return self._module_name

    def _error(self, msg):
        self._errors.append(msg)

    def _log_errors(self, errors, path):
        logging.error(path)
        for error in errors:
            logging.error('  %s' % error)


def _location(t, token_position):
    lineno = t.lineno(token_position)
    return pdefc.lang.Location(lineno)


def _with_location(token_position):
    def decorator(func):
        def set_location(self, t):
            func(self, t)
            t[0].location = _location(t, token_position)

        functools.update_wrapper(set_location, func)
        return set_location

    return decorator


class _Tokens(object):
    '''Lexer tokens.'''

    # Simple reserved words.
    types = (
        'BOOL',
        'INT16',
        'INT32',
        'INT64',
        'FLOAT',
        'DOUBLE',
        'STRING',
        'DATETIME',
        'VOID',

        'LIST',
        'SET',
        'MAP',

        'ENUM',
        'MESSAGE',
        'EXCEPTION',
        'INTERFACE')

    # Identifier types, see t_IDENTIFIER
    ids = types + ('FROM', 'IMPORT')
    ids_map = {s.lower(): s for s in ids}
    reserved = set(s.lower() for s in RESERVED)

    tokens = ids + (
        'DOT',
        'COLON',
        'COMMA',
        'SEMI',
        'LESS',
        'GREATER',
        'LBRACE',
        'RBRACE',
        'LPAREN',
        'RPAREN',
        'IDENTIFIER',
        'DOC',
        'DISCRIMINATOR',
        'POST',
        'QUERY',
        'THROWS')

    # Regexp for simple rules.
    t_DOT = r'.'
    t_COLON = r'\:'
    t_COMMA = r'\,'
    t_SEMI = r'\;'
    t_LESS = r'\<'
    t_GREATER = r'\>'
    t_LBRACE = r'\{'
    t_RBRACE = r'\}'
    t_LPAREN = r'\('
    t_RPAREN = r'\)'

    # Regexp for options.
    t_DISCRIMINATOR = r'@discriminator'
    t_POST = r'@post'
    t_QUERY = r'@query'
    t_THROWS = r'@throws'

    # Ignored characters
    t_ignore = " \t"

    doc_start_pattern = re.compile('^\s*\**\s*', re.MULTILINE)
    doc_end_pattern = re.compile('\s\*$', re.MULTILINE)

    def t_IDENTIFIER(self, t):
        r'[a-zA-Z_]{1}[a-zA-Z0-9_]*'
        t.type = self.ids_map.get(t.value, 'IDENTIFIER')

        if t.value.lower() in self.reserved:
            self._error('Line %s, "%s" is a reserved word' % (t.lineno, t.value))

        return t

    def t_comment(self, t):
        r'//.*\n'
        t.lexer.lineno += 1
        t.lineno += 1

    # Skip a new line and increment the lexer lineno counter.
    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += t.value.count("\n")

    # Pdef docstring.
    def t_DOC(self, t):
        r'\/\*\*((.|\n)*?)\*\/'
        t.lexer.lineno += t.value.count('\n')

        value = t.value.strip('/')
        value = self.doc_start_pattern.sub('', value)
        value = self.doc_end_pattern.sub('', value)
        t.value = value
        return t

    def t_error(self, t):
        self._error("Illegal character '%s', line %s" % (t.value[0], t.lexer.lineno))
        t.lexer.skip(1)

    def _error(self, msg):
        raise NotImplementedError()


class _GrammarRules(object):
    '''Parser grammar rules.'''
    def _name(self):
        raise NotImplementedError

    def _error(self, msg):
        raise NotImplementedError

    # Starting point.
    @_with_location(0)
    def p_module(self, p):
        '''
        module : doc imports definitions
        '''
        name = self._name()
        doc = p[1]
        imports = p[2]
        definitions = p[3]
        p[0] = pdefc.lang.Module(name, imports=imports, definitions=definitions, doc=doc)

    # Any absolute name, returns a list.
    def p_absolute_name(self, p):
        '''
        absolute_name : absolute_name DOT IDENTIFIER
                      | IDENTIFIER
        '''
        self._list(p, separated=True)

    def p_type_name(self, p):
        '''
        type_name : absolute_name
        '''
        p[0] = '.'.join(p[1])

    def p_module_name(self, p):
        '''
        module_name : absolute_name
        '''
        p[0] = '.'.join(p[1])

    # Empty token to support optional values.
    def p_empty(self, p):
        '''
        empty :
        '''
        pass

    def p_doc(self, p):
        '''
        doc : DOC
            | empty
        '''
        if len(p) == 2:
            p[0] = p[1]
        else:
            p[0] = ''

    def p_imports(self, p):
        '''
        imports : imports import
                | import
                | empty
        '''
        self._list(p)

    @_with_location(1)
    def p_import(self, p):
        '''
        import : absolute_import
               | relative_import
        '''
        p[0] = p[1]

    def p_absolute_import(self, p):
        '''
        absolute_import : IMPORT module_name SEMI
        '''
        p[0] = pdefc.lang.AbsoluteImport(p[2])

    def p_relative_import(self, p):
        '''
        relative_import : FROM module_name IMPORT relative_import_names SEMI
        '''
        p[0] = pdefc.lang.RelativeImport(p[2], p[4])

    def p_relative_import_names(self, p):
        '''
        relative_import_names : relative_import_names COMMA module_name
                              | module_name
        '''
        self._list(p, separated=True)

    def p_definitions(self, p):
        '''
        definitions : definitions definition
                    | definition
                    | empty
        '''
        self._list(p)

    def p_definition(self, p):
        '''
        definition : doc enum
                   | doc message
                   | doc interface
        '''
        d = p[2]
        d.doc = p[1]
        p[0] = d

    @_with_location(2)
    def p_enum(self, p):
        '''
        enum : ENUM IDENTIFIER LBRACE enum_values RBRACE
        '''
        p[0] = pdefc.lang.Enum(p[2], values=p[4])

    def p_enum_values(self, p):
        '''
        enum_values : enum_value_list SEMI
                    | enum_value_list
        '''
        p[0] = p[1]

    def p_enum_value_list(self, p):
        '''
        enum_value_list : enum_value_list COMMA enum_value
                        | enum_value
                        | empty
        '''
        self._list(p, separated=1)

    @_with_location(1)
    def p_enum_value(self, p):
        '''
        enum_value : IDENTIFIER
        '''
        p[0] = pdefc.lang.EnumValue(p[1])

    # Message definition
    @_with_location(3)
    def p_message(self, p):
        '''
        message : message_or_exc IDENTIFIER message_base LBRACE fields RBRACE
        '''
        is_exception = p[1].lower() == 'exception'
        name = p[2]
        base, discriminator_value = p[3]
        fields = p[5]

        p[0] = pdefc.lang.Message(name, base=base, discriminator_value=discriminator_value,
                                 declared_fields=fields, is_exception=is_exception)

    def p_message_or_exception(self, p):
        '''
        message_or_exc : MESSAGE
                       | EXCEPTION
        '''
        p[0] = p[1]

    def p_message_base(self, p):
        '''
        message_base : COLON type LPAREN type RPAREN
                     | COLON type
                     | empty
        '''
        base, discriminator = None, None

        if len(p) == 3:
            base = p[2]
        elif len(p) == 6:
            base = p[2]
            discriminator = p[4]

        if base:
            base.location = _location(p, 2)

        p[0] = base, discriminator

    # List of message fields
    def p_fields(self, p):
        '''
        fields : fields field
               | field
               | empty
        '''
        self._list(p)

    # Single message field
    @_with_location(1)
    def p_field(self, p):
        '''
        field : IDENTIFIER type field_discriminator SEMI
        '''
        name = p[1]
        type0 = p[2]
        is_discriminator = p[3]
        p[0] = pdefc.lang.Field(name, type0, is_discriminator=is_discriminator)

    def p_field_discriminator(self, p):
        '''
        field_discriminator : DISCRIMINATOR
                            | empty
        '''
        p[0] = bool(p[1])

    # Interface definition
    @_with_location(3)
    def p_interface(self, p):
        '''
        interface : interface_exc INTERFACE IDENTIFIER LBRACE methods RBRACE
        '''
        exc = p[1]
        name = p[3]
        methods = p[5]

        p[0] = pdefc.lang.Interface(name, exc=exc, methods=methods)

    def p_interface_exc(self, p):
        '''
        interface_exc : THROWS LPAREN type RPAREN
                      | empty
        '''
        if len(p) == 5:
            p[0] = p[3]
        else:
            p[0] = None

    def p_methods(self, p):
        '''
        methods : methods method
                | method
                | empty
        '''
        self._list(p)

    @_with_location(3)
    def p_method(self, p):
        '''
        method : doc method_post IDENTIFIER LPAREN method_args RPAREN type SEMI
        '''
        doc = p[1]
        is_post = p[2]
        name = p[3]
        args = p[5]
        result = p[7]
        p[0] = pdefc.lang.Method(name, result=result, args=args, is_post=is_post, doc=doc)

    def p_method_post(self, p):
        '''
        method_post : POST
                    | empty
        '''
        p[0] = p[1] == '@post'

    def p_method_args(self, p):
        '''
        method_args : method_args COMMA method_arg
                    | method_arg
                    | empty
        '''
        self._list(p, separated=True)

    @_with_location(2)
    def p_method_arg(self, p):
        '''
        method_arg : doc IDENTIFIER type method_arg_attr
        '''
        attr = p[4]
        is_query = attr == '@query'
        is_post = attr == '@post'
        p[0] = pdefc.lang.MethodArg(p[2], p[3], is_query=is_query, is_post=is_post)

    def p_method_arg_attr(self, p):
        '''
        method_arg_attr : POST
                        | QUERY
                        | empty
        '''
        p[0] = p[1]

    @_with_location(1)
    def p_type(self, p):
        '''
        type : value_ref
             | list_ref
             | set_ref
             | map_ref
             | def_ref
        '''
        p[0] = p[1]

    def p_value_ref(self, p):
        '''
        value_ref : BOOL
                  | INT16
                  | INT32
                  | INT64
                  | FLOAT
                  | DOUBLE
                  | STRING
                  | DATETIME
                  | VOID
        '''
        p[0] = pdefc.lang.reference(p[1].lower())

    def p_list_ref(self, p):
        '''
        list_ref : LIST LESS type GREATER
        '''
        p[0] = pdefc.lang.ListReference(p[3])

    def p_set_ref(self, p):
        '''
        set_ref : SET LESS type GREATER
        '''
        p[0] = pdefc.lang.SetReference(p[3])

    def p_map_ref(self, p):
        '''
        map_ref : MAP LESS type COMMA type GREATER
        '''
        p[0] = pdefc.lang.MapReference(p[3], p[5])

    def p_def_ref(self, p):
        '''
        def_ref : type_name
        '''
        p[0] = pdefc.lang.reference(p[1])

    def p_error(self, p):
        if p is None:
            msg = 'Unexpected end of file'
        else:
            msg = 'Line %s, syntax error at "%s"' % (p.lexer.lineno, p.value)

        self._error(msg)

    def _list(self, p, separated=False):
        '''List builder, supports separated and empty lists.

        Supported grammar:
        list : list [optional separator] item
             | item
             | empty
        '''
        if len(p) == 1:
            p[0] = []
        elif len(p) == 2:
            if p[1] is None:
                p[0] = []
            else:
                p[0] = [p[1]]
        else:
            p[0] = p[1]
            if not separated:
                p[0].append(p[2])
            else:
                p[0].append(p[3])


class _Grammar(_GrammarRules, _Tokens):
    '''Grammar combines grammar rules and lexer tokens. It can be passed to the ply.yacc
    and pla.lex functions as the module argument.'''

    def __init__(self, name_func, error_func):
        self._name = name_func
        self._error = error_func



# encoding: utf-8
import logging
from pdefc.lang.types import Type, TypeEnum, Definition


class Enum(Definition):
    '''Enum definition.'''
    def __init__(self, name, values=None, value_names=None, doc=None, location=None):
        super(Enum, self).__init__(TypeEnum.ENUM, name, doc=doc, location=location)
        self.values = []

        if values and value_names:
            raise ValueError('values and value_names are mutually exclusive')

        if values:
            for value in values:
                self.add_value(value)

        if value_names:
            for name in value_names:
                self.create_value(name)

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<%s %s %s>' % (self.__class__.__name__, self.name, hex(id(self)))

    def add_value(self, value):
        '''Add a value to this enum.'''
        self.values.append(value)
        logging.debug('Added an enum value, enum=%s, value=%s' % (self, value))

    def create_value(self, name):
        '''Create a new enum value by its name, add it to this enum, and return it.'''
        value = EnumValue(name, enum=self)
        self.add_value(value)
        return value

    def get_value(self, name):
        '''Get a value by its name or raise an exception.'''
        for value in self.values:
            if value.name == name:
                return value

    def __contains__(self, enum_value):
        return enum_value in self.values

    def link(self, module):
        errors = super(Enum, self).link(module)

        for value in self.values:
            errors += value.link(self)

        return errors

    def _validate(self):
        errors = []

        names = set()
        for value in self.values:
            if value.name in names:
                errors.append(self._error('%s: duplicate enum value "%s"', self, value.name))

            names.add(value.name)

        return errors


class EnumValue(Type):
    '''Single enum value which has a name and a pointer to the declaring enum.'''
    def __init__(self, name, enum=None, location=None):
        super(EnumValue, self).__init__(TypeEnum.ENUM_VALUE, location=location)
        self.name = name
        self.enum = enum

    def __str__(self):
        if not self.enum:
            return self.name
        return self.enum.name + '.' + self.name

    def __repr__(self):
        return '<%s %s at %s>' % (self.__class__.__name__, self.name, hex(id(self)))

    def link(self, enum):
        if self.enum and not self.enum is enum:
            raise ValueError('Enum value is already linked, value=%s' % self)
        self.enum = enum
        return []



# encoding: utf-8
import logging
from collections import deque
import re

from pdefc.lang.common import Validatable
from pdefc.lang.types import NativeType, TypeEnum


class Module(Validatable):
    '''Module is a named scope for definitions. It is usually a *.pdef file.'''
    name_pattern = re.compile(r'^[a-zA-Z]{1}[a-zA-Z0-9_]*(\.[a-zA-Z]{1}[a-zA-Z0-9_]*)*$')

    def __init__(self, relative_name, imports=None, definitions=None, doc=None, path=None):
        self.relative_name = relative_name
        self.doc = doc
        self.path = path
        self.package = None

        self.imports = []               # imports
        self.imported_aliases = []      # (alias, module) pairs
        self.imported_modules = []      # modules

        self.definitions = []
        self._definition_map = {}       # Performance optimization

        if imports:
            for import0 in imports:
                self.add_import(import0)

        if definitions:
            for def0 in definitions:
                self.add_definition(def0)

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<%s %s at %s>' % (self.__class__.__name__, self.name, hex(id(self)))

    @property
    def name(self):
        if not self.package:
            return self.relative_name

        if self.package.name == self.relative_name:
            return self.relative_name

        return '%s.%s' % (self.package.name, self.relative_name)

    @property
    def imported_definitions(self):
        '''Return a set of imported definitions.'''
        result = set()
        for def0 in self.definitions:
            result.update(def0.imported_definitions)
        
        return result

    def _log_return_errors(self, errors):
        if not errors:
            return []

        logging.error(self.path or self.name)
        for error in errors:
            logging.error('  %s' % error)

        return errors

    def add_import(self, import0):
        '''Add a module import to this module.'''
        self.imports.append(import0)
        logging.debug('%s: added an import "%s"', self, import0)

    def add_imported_module(self, alias, module):
        '''Add an imported module to this module.'''
        self.imported_aliases.append((alias, module))
        self.imported_modules.append(module)

        logging.debug('%s: added an imported module, alias="%s", module="%s"', self, alias, module)

    def get_imported_module(self, alias):
        '''Find a module by its import alias.'''
        for alias1, module in self.imported_aliases:
            if alias1 == alias:
                return module

        return None

    def add_definition(self, def0):
        '''Add a new definition to this module.'''
        self.definitions.append(def0)
        self._definition_map[def0.name] = def0
        logging.debug('%s: added a definition "%s"', self, def0.name)

    def get_definition(self, name):
        '''Return a definition or an enum value in this module by a name.'''

        # Try to get a definition by name.
        if name in self._definition_map:
            return self._definition_map[name]

        # Return if not a relative name.
        if '.' not in name:
            return

        # Try to get an enum value.
        enum_name, value_name = name.split('.', 1)
        if enum_name not in self._definition_map:
            return

        enum = self._definition_map[enum_name]
        if not enum.is_enum:
            return

        return enum.get_value(value_name)

    def lookup(self, name):
        '''Find a definition or enum value in this module or imported modules.'''

        # Try to get a native type.
        def0 = NativeType.get(name)
        if def0:
            return def0

        # Try to find a type or an enum value in the current module.
        def0 = self.get_definition(name)
        if def0:
            return def0

        # Try to find an imported type.
        left = []
        right = name.split('.')
        while right:
            left.append(right.pop(0))
            lname = '.'.join(left)
            rname = '.'.join(right)

            module = self.get_imported_module(lname)
            if not module:
                continue

            # Try to get a definition or an enum value from the imported module.
            def0 = module.get_definition(rname)
            if def0:
                return def0

        return None

    def _depends_on(self, another):
        return bool(self._get_import_path(another))

    def _get_import_path(self, another):
        '''Returns a list of modules when this module depends on another module.

        For example: a imports b, b imports c;
        a._get_import_path_with(c) => [a, b, c]
        '''
        if another is None or another is self:
            return []

        # BFS
        q = deque(self.imported_modules)
        seen = set()
        came_from = {module: self for module in self.imported_modules}

        depends = False
        while q:
            current = q.pop()
            if current in seen:
                continue

            if current is another:
                # Self depends on another!
                depends = True
                break

            seen.add(current)
            q += current.imported_modules
            came_from.update({imported: current for imported in current.imported_modules})

        if not depends:
            return []

        # Reconstruct the path.
        path = [another]
        module = another
        while True:
            module = came_from[module]
            path.append(module)
            if module is self:
                break

        path.reverse()
        return path

    # Link.

    def link(self, package=None):
        '''Link imports and definitions and return a list of errors.'''
        logging.debug('Linking %s as %s', self.path, self)

        if self.package:
            raise ValueError('Module is already linked, module=%s' % self)
        self.package = package

        errors = self._link_imports(package)
        if errors:
            return self._log_return_errors(errors)

        errors = self._link_definitions()
        return self._log_return_errors(errors)

    def _link_imports(self, package):
        '''Link imports, must be called before link_module_defs().'''
        errors = []

        for import0 in self.imports:
            errors += import0.link(package)

            for alias, module in import0.alias_module_pairs:
                self.add_imported_module(alias, module)

        return errors

    def _link_definitions(self):
        '''Link imports and definitions.'''
        errors = []
        for def0 in self.definitions:
            errors += def0.link(self)

        return errors

    # Build.

    def build(self):
        '''Build definitions and return a list of errors.'''
        logging.debug('Building %s', self)

        errors = []
        for def0 in self.definitions:
            errors += def0.build()

        return self._log_return_errors(errors)

    # Validate.

    def _validate(self):
        '''Validate imports and definitions and return a list of errors.'''
        logging.debug('Validating %s', self)
        errors = []
        errors += self._validate_name()
        errors += self._validate_no_duplicate_symbols()

        for def0 in self.definitions:
            errors += def0.validate()

        return self._log_return_errors(errors)

    def _validate_name(self):
        if self.name_pattern.match(self.relative_name):
            return []
        return ['Wrong module name "%s". A name must be one or several words separated by dots, '
                'a word must contain only latin letters, digits and underscores, '
                'and must start with a letter, for example, "users.accounts.events"' % self.name]

    def _validate_no_duplicate_symbols(self):
        errors = []

        # Prevent imports with duplicate aliases.
        names = set()
        for alias, _ in self.imported_aliases:
            if alias in names:
                errors.append('Duplicate import "%s"' % alias)
            names.add(alias)

        # Prevent definitions and imports with duplicate names.
        for def0 in self.definitions:
            name = def0.name
            if name in names:
                errors.append('Duplicate definition or import "%s"' % name)
            names.add(name)

        return errors



# encoding: utf-8
import inspect
import io
import logging
import os
import pkg_resources
from jinja2 import Environment

GENERATORS_ENTRY_POINT = 'pdefc.generators'
ENCODING = 'utf8'


class Generator(object):
    '''Abstract code generator, subclass it and implement the generate method.

    @param out          generated files destination (directory).
    @param namespace    an optional {'pdef.package.module': 'language_module'} mapping.
    '''
    def __init__(self, out, namespace=None, **kwargs):
        self.out = out
        self.namespace = Namespace(namespace)

    def generate(self, package):
        raise NotImplementedError

    def write_file(self, filename, text):
        '''Write a text file to the output directory, filename can contain subdirectories.'''
        # Join the filename with the destination directory.
        filepath = os.path.join(self.out, filename)

        # Create a directory with its children for a file.
        dirpath = os.path.dirname(filepath)
        mkdir_p(dirpath)

        # Write the file contents.
        with io.open(filepath, 'wt', encoding=ENCODING) as f:
            f.write(text)
        logging.info('Created %s', filepath)


def find_generators(entry_point=GENERATORS_ENTRY_POINT):
    '''Dynamically load the source code generators, return a dict {name: generator}.'''
    result = {}
    for entry_point in pkg_resources.iter_entry_points(group=entry_point):
        name = entry_point.name
        generator = entry_point.load()
        result[name] = generator
        logging.debug('Loaded a source code generator %r', name)

    return result


class Templates(object):
    '''Templates class is a default Jinja templates loader relative to a directory or a file.

    Get a template::
        >>> templates = Templates(__file__)
        >>> templates.get('my_jinja.jinja2')

    Render a template::
        >>> templates = Templates(__file__)
        >>> templates.render('mytemplate.jinja2', key='value')

    '''
    def __init__(self, dir_or_file, filters=None):
        '''Create a templates loader relative to a directory or a file.'''
        if os.path.isdir(dir_or_file):
            self._dir = dir_or_file
        else:
            self._dir = os.path.dirname(dir_or_file)

        self._env = Environment(trim_blocks=True, lstrip_blocks=True)
        self._cache = {}

        self.add_filter('upper_first', upper_first)
        if isinstance(filters, dict):
            self.add_filters(**filters)
        elif filters:
            self.add_filters_from_methods(filters)

    def add_filter(self, name, filter0):
        self._env.filters[name] = filter0
        logging.debug('Added a template filter "%s"' % name)

    def add_filters(self, **name_to_filter):
        for name, filter0 in name_to_filter.items():
            self.add_filter(name, filter0)

    def add_filters_from_methods(self, obj):
        for name in dir(obj):
            if name.startswith('_'):
                continue

            attr = getattr(obj, name)
            if inspect.ismethod(attr):
                self.add_filter(name, attr)

    def get(self, name):
        '''Read and return a Jinja template, the templates are cached.'''
        if name in self._cache:
            return self._cache[name]

        # Get the template file.
        path = os.path.join(self._dir, name)
        with open(path, 'r') as module_file:
            text = module_file.read()

        template = self._env.from_string(text)
        self._cache[name] = template

        return template

    def render(self, template_name, **kwargs):
        '''Get a template and render it using the keyword arguments.'''
        template = self.get(template_name)
        return template.render(**kwargs)


class Namespace(object):
    '''Namespace class is a default namespace mapper, which maps pdef modules names to
    language module names.

    Example::
        >>> namespace = Namespace({'pdef.tests': 'pdef_tests'})
        >>> namespace.map('pdef.tests.messages')
        >>> 'pdef_tests.messages'

    '''
    def __init__(self, namespace=None):
        self._namespace = dict(namespace) if namespace else {}

    def __call__(self, module_name):
        return self.map(module_name)

    def map(self, name):
        '''Returns a new name.'''
        for prefix, mapped in self._namespace.items():
            if name == prefix:
                # Full match, service.module => service_module.
                return mapped

            if name.startswith(prefix + '.'):
                return mapped + name[len(prefix):]

        return name


def upper_first(s):
    '''Uppercase the first letter in a string.'''
    if not s:
        return s
    return s[0].upper() + s[1:]


def mkdir_p(dirname):
    '''Make directories, ignore errors'''
    if os.path.exists(dirname):
        return
    os.makedirs(dirname)



# encoding: utf-8
from pdefc.lang.common import Located, Validatable


class TypeEnum(object):
    '''TypeEnum is an enumeration of all pdef type tokens.'''

    # Base value types.
    BOOL = 'bool'
    INT16 = 'int16'
    INT32 = 'int32'
    INT64 = 'int64'
    FLOAT = 'float'
    DOUBLE = 'double'
    STRING = 'string'
    DATETIME = 'datetime'

    # Collection types.
    LIST = 'list'
    MAP = 'map'
    SET = 'set'

    # User defined data types.
    ENUM = 'enum'
    ENUM_VALUE = 'enum_value'
    MESSAGE = 'message'

    # Interface and void.
    INTERFACE = 'interface'
    VOID = 'void'

    PRIMITIVE_TYPES = (BOOL, INT16, INT32, INT64, FLOAT, DOUBLE, STRING, DATETIME)
    DATA_TYPES = PRIMITIVE_TYPES + (LIST, SET, MAP, ENUM, MESSAGE)
    NATIVE_TYPES = PRIMITIVE_TYPES + (VOID, )
    COLLECTION_TYPES = (LIST, SET, MAP)
    DEFINITION_TYPES = (ENUM, MESSAGE, INTERFACE)

    @classmethod
    def is_primitive(cls, type_enum):
        return type_enum in cls.PRIMITIVE_TYPES

    @classmethod
    def is_data_type(cls, type_enum):
        return type_enum in cls.DATA_TYPES

    @classmethod
    def is_collection(cls, type_enum):
        return type_enum in cls.COLLECTION_TYPES


class Type(Located, Validatable):
    '''Type is a common class for all pdef types. These include native types, definitions,
    collections, and enum values.'''
    def __init__(self, type0, location=None):
        self.type = type0
        self.location = location
        self.is_exception = False

    @property
    def is_primitive(self):
        return self.type in TypeEnum.PRIMITIVE_TYPES

    @property
    def is_data_type(self):
        return self.type in TypeEnum.DATA_TYPES

    @property
    def is_native(self):
        return self.type in TypeEnum.NATIVE_TYPES

    @property
    def is_collection(self):
        return TypeEnum.is_collection(self.type)

    @property
    def is_definition(self):
        return self.type in TypeEnum.DEFINITION_TYPES

    @property
    def is_message(self):
        return self.type == TypeEnum.MESSAGE

    @property
    def is_interface(self):
        return self.type == TypeEnum.INTERFACE

    @property
    def is_enum(self):
        return self.type == TypeEnum.ENUM

    @property
    def is_enum_value(self):
        return self.type == TypeEnum.ENUM_VALUE

    def __str__(self):
        return self.type

    def __repr__(self):
        return '<%s %s at %s>' % (self.__class__.__name__, self.type, hex(id(self)))


class NativeType(object):
    '''Native type singletons.'''
    BOOL = Type(TypeEnum.BOOL)
    INT16 = Type(TypeEnum.INT16)
    INT32 = Type(TypeEnum.INT32)
    INT64 = Type(TypeEnum.INT64)
    FLOAT = Type(TypeEnum.FLOAT)
    DOUBLE = Type(TypeEnum.DOUBLE)
    STRING = Type(TypeEnum.STRING)
    DATETIME = Type(TypeEnum.DATETIME)
    VOID = Type(TypeEnum.VOID)

    _BY_TYPE = {
        TypeEnum.BOOL: BOOL,
        TypeEnum.INT16: INT16,
        TypeEnum.INT32: INT32,
        TypeEnum.INT64: INT64,
        TypeEnum.FLOAT: FLOAT,
        TypeEnum.DOUBLE: DOUBLE,
        TypeEnum.STRING: STRING,
        TypeEnum.DATETIME: DATETIME,
        TypeEnum.VOID: VOID
    }

    @classmethod
    def get(cls, name):
        '''Return a value by its type or none.'''
        return cls._BY_TYPE.get(name)

    @classmethod
    def all(cls):
        '''Return all native types.'''
        return list(cls._BY_TYPE.values())


class Definition(Type):
    '''Definition is a base for user-specified types. These include messages,
    interfaces and enums.'''

    def __init__(self, type0, name, doc=None, location=None):
        super(Definition, self).__init__(type0, location=location)
        self.name = name
        self.doc = doc
        self.module = None

    def __repr__(self):
        return '<%s %s at %s>' % (self.__class__.__name__, self.name, hex(id(self)))

    def __str__(self):
        return self.name

    @property
    def referenced_types(self):
        '''Return a list of all types referenced in this definition (in fields, methods, etc).'''
        return []

    @property
    def imported_definitions(self):
        '''Return a set of referenced definitions which are in imported modules.'''
        result = set()

        # Filter imported definitions.
        for type0 in self.referenced_types:
            if type0.is_enum_value:
                type0 = type0.enum

            if not type0.is_definition:
                continue

            if type0.module == self.module:
                continue

            result.add(type0)

        return result

    def lookup(self, name):
        return self.module.lookup(name) if self.module else None

    def link(self, module):
        '''Link this definition references and return a list of errors.'''
        if self.module:
            raise ValueError('Definition is already linked')

        self.module = module
        return []

    def build(self):
        '''Build this definition after linking and return a list of errors.'''
        return []

    def _is_defined_after(self, another):
        '''Return true if this definition is after another one in one module.'''
        if not self.module or not another.module:
            return True

        if self.module is not another.module:
            return True

        for def0 in self.module.definitions:
            if def0 is another:
                return True

            if def0 is self:
                return False

        raise AssertionError('Unreachable code')



# encoding: utf-8
from pdefc.cli import main
from pdefc.version import __version__

from pdefc.exc import CompilerException
from pdefc.compiler import create_compiler
from pdefc.parser import create_parser
from pdefc.sources import create_sources
from pdefc.generators import find_generators



# encoding: utf-8
import logging

import pdefc.lang.collects
from pdefc.lang.common import Located, Validatable
from pdefc.lang.types import Type

try:
    # Python 2.7
    string = basestring
except NameError:
    # Python 3
    string = str


def reference(name_ref_def):
    '''Create a reference from a string name, another reference, a definition or None'''
    if name_ref_def is None:
        return EmptyReference()

    elif isinstance(name_ref_def, string):
        return NameReference(name_ref_def)

    elif isinstance(name_ref_def, Type):
        return Reference(name_ref_def)

    elif isinstance(name_ref_def, Reference):
        return name_ref_def

    raise ValueError('Unsupported type: %r' % name_ref_def)


class Reference(Located, Validatable):
    '''Reference directly references a type.'''
    def __init__(self, type0=None, location=None):
        self._type = type0
        self.location = location

    def __bool__(self):
        return bool(self._type)

    def __nonzero__(self):
        return bool(self._type)

    def dereference(self):
        '''Return a type this references points to or raise ValueError when not linked.'''
        if not self._type:
            raise ValueError('Reference is not linked: %s' % self)
        return self._type

    def link(self, lookup):
        '''Link this reference in a provided callable lookup.'''
        return []

    @property
    def referenced_types(self):
        '''Return a list of all referenced types in this reference.'''
        if not self._type:
            return []
        return [self._type]


class EmptyReference(Reference):
    '''EmptyReference is a sentinel for an absent type. It returns None when dereferenced'''
    def __init__(self, location=None):
        super(EmptyReference, self).__init__(None, location=location)

    def dereference(self):
        return None


class NameReference(Reference):
    '''NameReference references a type by its name.'''
    def __init__(self, name, location=None):
        super(NameReference, self).__init__(None, location=location)
        self.name = name

    def __repr__(self):
        return '<NameReference %r>' % self.name

    def __str__(self):
        return self.name

    def link(self, lookup):
        logging.debug('Linking %s', self)
        self._type = lookup(self.name)
        if self._type:
            return []

        return [self._error('Type not found %r', self.name)]


class ListReference(Reference):
    '''ListReference has a child reference for an element, creates a list on linking.'''
    def __init__(self, element, location=None):
        super(ListReference, self).__init__(None, location=location)
        self.element = reference(element)
        self._init_type()

    def __repr__(self):
        return '<ListReference %s>' % self.element

    def __str__(self):
        return 'list<%s>' % self.element

    def _init_type(self):
        if not self.element:
            return
        self._type = pdefc.lang.collects.List(self.element.dereference(), location=self.location)

    def link(self, lookup):
        logging.debug('Linking %s', self)
        errors = self.element.link(lookup)
        if errors:
            return errors

        self._init_type()
        return []

    def _validate(self):
        logging.debug('Validating %s', self)
        if not self._type:
            return []
        return self._type.validate()

    @property
    def referenced_types(self):
        result = super(ListReference, self).referenced_types
        result += self.element.referenced_types
        return result


class SetReference(Reference):
    '''SetReference has a child for an element, creates a set on linking.'''
    def __init__(self, element, location=None):
        super(SetReference, self).__init__(None, location=location)
        self.element = reference(element)
        self._init_type()

    def __repr__(self):
        return '<SetReference %s>' % self.element

    def __str__(self):
        return 'set<%s>' % self.element

    def _init_type(self):
        if not self.element:
            return
        self._type = pdefc.lang.collects.Set(self.element.dereference(), location=self.location)

    def link(self, lookup):
        logging.debug('Linking %s', self)
        errors = self.element.link(lookup)
        if errors:
            return errors

        self._init_type()
        return []

    def _validate(self):
        logging.debug('Validating %s', self)
        if not self._type:
            return []
        return self._type.validate()

    @property
    def referenced_types(self):
        result = super(SetReference, self).referenced_types
        result += self.element.referenced_types
        return result


class MapReference(Reference):
    '''MapReference has children references for a key and a value, creates a map on linking.'''
    def __init__(self, key, value, location=None):
        super(MapReference, self).__init__(None, location=location)
        self.key = reference(key)
        self.value = reference(value)
        self._init_type()

    def __repr__(self):
        return '<MapReference %s, %s>' % (self.key, self.value)

    def __str__(self):
        return 'map<%s, %s>' % (self.key, self.value)

    def _init_type(self):
        if not self.key or not self.value:
            return
        self._type = pdefc.lang.collects.Map(self.key.dereference(), self.value.dereference(),
                                            location=self.location)

    def link(self, lookup):
        logging.debug('Linking %s', self)
        errors0 = self.key.link(lookup)
        errors1 = self.value.link(lookup)
        if errors0 or errors1:
            return errors0 + errors1

        self._init_type()
        return []

    def _validate(self):
        logging.debug('Validation %s', self)
        if not self._type:
            return []
        return self._type.validate()

    @property
    def referenced_types(self):
        result = super(MapReference, self).referenced_types
        result += self.key.referenced_types
        result += self.value.referenced_types
        return result



# encoding: utf-8
import logging
import time

import pdefc
from pdefc.exc import CompilerException
from pdefc.lang.packages import Package


def create_compiler(paths=None, allow_duplicate_definitions=False):
    '''Creates a compiler, the compiler is reusable but not thread-safe.'''
    sources = pdefc.create_sources(paths)
    return Compiler(sources, allow_duplicate_definitions=allow_duplicate_definitions)


class Compiler(object):
    '''Compiler parses and compiles packages. It is reusable but not thread-safe.'''
    def __init__(self, sources=None, parser=None, allow_duplicate_definitions=False):
        self.sources = sources or pdefc.create_sources()
        self.parser = parser or pdefc.create_parser()
        self.allow_duplicate_definitions = allow_duplicate_definitions
        self._generators = None

    @property
    def generators(self):
        if self._generators is None:
            self._generators = dict(pdefc.find_generators())
        return self._generators

    def check(self, package_path):
        '''Compile a package from a path, and print a message if its correct.'''
        package = self.compile(package_path)
        logging.info('%s is valid', package.name)

    def compile(self, package_path):
        '''Compile a package from a path.'''
        t0 = time.time()
        source = self.sources.read_path(package_path)
        package = self._compile(source.name, compiled_map=set())

        t1 = time.time()
        t = (t1 - t0) * 1000
        logging.info('Fetched and compiled %s in %dms', package.name, t)
        return package

    def generate(self, package_path, generator_name, out, namespace=None):
        '''Generate a package from a path.'''

        factory = self.generators.get(generator_name)
        if not factory:
            raise CompilerException('Source code generator not found: %s' % generator_name)

        package = self.compile(package_path)
        namespace = namespace or {}

        t0 = time.time()
        generator = factory(out, namespace=namespace)
        generator.generate(package)

        t1 = time.time()
        t = (t1 - t0) * 1000
        logging.info('Generated %s code in %dms', generator_name, t)

    def _compile(self, package_name, compiled_map):
        '''Compile and return a package with its dependencies, prevent circular dependencies.'''

        # Prevent circular package dependencies
        if package_name in compiled_map:
            raise CompilerException('Circular dependency: package=%s' % package_name)
        compiled_map.add(package_name)

        # Parse the package before compiling its dependencies
        # to improve responsiveness.
        package = self._parse(package_name)

        # Compile the package dependencies.
        deps = []
        for depname in package.info.dependencies:
            dependency = self._compile(depname, compiled_map)
            deps.append(dependency)

        # Compile and return the package.
        logging.info('Compiling %s', package_name)
        for dep in deps:
            package.add_dependency(dep)

        errors = package.compile(allow_duplicate_definitions=self.allow_duplicate_definitions)
        if errors:
            raise CompilerException('Compilation errors', errors)

        return package

    def _parse(self, pname):
        '''Parse and return a package, but not its dependencies.'''
        # Find the package source.
        source = self.sources.get(pname)
        info = source.info

        # Init the package.
        logging.info('Parsing %s', source)
        package = Package(pname, info=info)
        assert info.name == pname, 'Package name does not match one in info'

        # Parse package modules.
        errors = []
        modules = []
        for mname in info.modules:
            s = source.module(mname)
            path = source.module_path(mname)

            module, errors0 = self.parser.parse(s, relative_name=mname, path=path)
            errors += errors0
            if module:
                modules.append(module)

        if errors:
            raise CompilerException('Parsing errors', errors)

        for module in modules:
            package.add_module(module)
        return package



# encoding: utf-8
from pdefc.lang.collects import List, Set, Map
from pdefc.lang.common import Location
from pdefc.lang.types import Type, TypeEnum, Definition, NativeType
from pdefc.lang.enums import Enum, EnumValue
from pdefc.lang.imports import AbstractImport, AbsoluteImport, RelativeImport
from pdefc.lang.interfaces import Interface, Method, MethodArg
from pdefc.lang.messages import Message, Field
from pdefc.lang.modules import Module
from pdefc.lang.references import reference, ListReference, SetReference, MapReference



# encoding: utf-8
'''Reserved tokens.'''

RESERVED = (
    "abstract",
    "alias",
    "and",
    "as",
    "begin",
    "break",
    "byte",
    "case",
    "catch",
    "class",
    "clone",
    "const",
    "continue",
    "declare",
    "def",
    "default",
    "defined",
    "descriptor",
    "descriptors",
    "do",
    "elif",
    "else",
    "elseif",
    "end",
    "enddeclare",
    "endfor",
    "endforeach",
    "endif",
    "endswitch",
    "endwhile",
    "ensure",
    "extends",
    "false",
    "final",
    "for",
    "foreach",
    "function",
    "global",
    "goto",
    "if",
    "implements",
    "in",
    "instanceof",
    "int",
    "long",
    "module",
    "namespace",
    "new",
    "next",
    "nil",
    "not",
    "or",
    "private",
    "protected",
    "public",
    "redo",
    "rescue",
    "retry",
    "return",
    "self",
    "short",
    "static",
    "super",
    "switch",
    "then",
    "throw",
    "true",
    "try",
    "undef",
    "unless",
    "until",
    "use",
    "var",
    "when",
    "while",
    "xor",
    "yield"
)



# encoding: utf-8
import io
import logging
import os
import sys

try:
    # Python2.7
    import urllib2
    from urllib2 import urlopen
    from urlparse import urlparse, urljoin
except ImportError:
    # Python3+
    from urllib.request import urlopen
    from urllib.parse import urlparse, urljoin

from pdefc import CompilerException
from pdefc.lang.packages import PackageInfo


MODULE_EXT = '.pdef'
PACKAGE_EXT = '.yaml'
MODULE_SEP = '/'
UTF8 = 'utf-8'


def create_sources(paths=None):
    return Sources(paths)


class Sources(object):
    '''Package sources.'''
    def __init__(self, paths=None):
        self._sources = {}  # Package names to sources.
        self._files = {}    # Absolute file names to sources.

        if paths:
            for path in paths:
                self.add_path(path)

    def _create_file_source(self, filename):
        return FileSource(filename)

    def _create_url_source(self, url):
        return UrlSource(url)

    def add_source(self, source):
        '''Add a package source, raise CompilerException if a duplicate package.'''
        name = source.name
        if name in self._sources:
            raise CompilerException('Duplicate package source "%s"' % source)

        self._sources[name] = source
        logging.debug('Added a source "%s"' % source)
        return source

    def add_path(self, path):
        '''Add a source path.'''
        if os.path.isdir(path):
            return self._walk_directory(path)
        return self.read_path(path)

    def read_path(self, path):
        '''Read a package source path, add and return the source.'''
        if os.path.isfile(path):
            return self.read_file(path)
        elif _is_url(path):
            return self.read_url(path)

        raise CompilerException('Unsupported path: %s' % path)

    def read_url(self, url):
        '''Read a package source url, add and return the source.'''
        source = self._create_url_source(url)
        self.add_source(source)
        return source

    def read_file(self, filename):
        '''Read a package source file, add and return the source.'''
        absolute = os.path.abspath(filename)
        if absolute in self._files:
            logging.debug('Package file is already present: %s', absolute)
            return self._files[absolute]

        source = self._create_file_source(filename)
        self._files[absolute] = source
        return self.add_source(source)

    def get(self, package_name):
        '''Return a package source by a package name.'''
        source = self._sources.get(package_name)
        if not source:
            raise CompilerException('Package not found "%s"' % package_name)

        return source

    def _walk_directory(self, dirname, package_ext=PACKAGE_EXT):
        if not os.path.exists(dirname):
            raise CompilerException('Directory does not exist: %s' % dirname)

        if not os.path.isdir(dirname):
            raise CompilerException('Not a directory: %s' % dirname)

        result = {}
        for root, dirs, files in os.walk(dirname):
            for file0 in files:
                ext = os.path.splitext(file0)[1]
                if ext.lower() != package_ext:
                    continue

                filename = os.path.join(root, file0)
                self.read_file(filename)

        return result


class Source(object):
    '''Package source.'''
    name = None     # Package name
    info = None     # Package info

    def module(self, module_name):
        '''Return a module source.'''
        raise NotImplementedError

    def module_path(self, module_name):
        '''Return a debug module path which can be used in errors, logs, etc.'''
        raise NotImplementedError

    def _module_filename(self, module_name, ext=MODULE_EXT, sep=MODULE_SEP):
        return module_name.replace('.', sep) + ext


class InMemorySource(Source):
    def __init__(self, info, modules_to_sources=None):
        self.name = info.name
        self.info = info
        self.modules = dict(modules_to_sources) if modules_to_sources else {}

    def module(self, module_name):
        return self.modules.get(module_name)

    def module_path(self, module_name):
        return self._module_filename(module_name)


class FileSource(Source):
    def __init__(self, filename):
        self.filename = filename
        self.dirname = os.path.dirname(filename)
        s = self._read(filename)

        try:
            self.info = PackageInfo.from_yaml(s)
            self.name = self.info.name
        except Exception as e:
            raise CompilerException('Failed to read package info: %s, e=%s' % (filename, e))

    def __str__(self):
        return self.filename

    def __repr__(self):
        return '<FileSource %r, file=%r at %s>' % (self.name, self.filename, hex(id(self)))

    def module(self, module_name):
        filepath = self.module_path(module_name)
        return self._read(filepath)

    def module_path(self, module_name):
        filename = self._module_filename(module_name)
        return os.path.join(self.dirname, filename)

    def _read(self, filepath):
        if not os.path.exists(filepath):
            raise CompilerException('File does not exist: %s' % filepath)

        try:
            with io.open(filepath, 'r', encoding=UTF8) as f:
                return f.read()
        except Exception as e:
            raise CompilerException('Failed to read a file: %s, e=%s' % (filepath, e))


class UrlSource(Source):
    def __init__(self, url):
        self.url = url
        self._delegate = None

    def __str__(self):
        if not self._delegate:
            return self.url
        return self.url

    def __repr__(self):
        return '<UrlSource url=%r at %s>' % (self.url, hex(id(self)))

    @property
    def name(self):
        return self.delegate().name

    @property
    def info(self):
        return self.delegate().info

    def module(self, module_name):
        return self.delegate().module(module_name)

    def module_path(self, module_name):
        filename = self._module_filename(module_name)
        return urljoin(self.url, filename)

    def delegate(self):
        if self._delegate:
            return self._delegate

        package_json = self._fetch(self.url)
        info = PackageInfo.from_yaml(package_json)

        modules = {}
        for name in info.modules:
            url = self.module_path(name)
            modules[name] = self._fetch(url)

        self._delegate = InMemorySource(info, modules)
        return self._delegate

    def _fetch(self, url):
        logging.warn('Downloading %s', url)
        try:
            req = urlopen(url)
            data = req.read()
            if sys.version < '3':
                return data
            return data.decode()
        except Exception as e:
            raise CompilerException('%s: %s' % (e, url))


def _is_url(s):
    if not s:
        return False

    scheme = urlparse(s).scheme
    return scheme and scheme.lower() in ('http', 'https', 'ftp')



# encoding: utf-8
import logging


class Location(object):
    def __init__(self, lineno):
        self.lineno = lineno

    def __str__(self):
        return 'line %s' % self.lineno

    def __repr__(self):
        return '<%s %s at %s>' % (self.__class__.__name__, self.lineno, hex(id(self)))

    def __eq__(self, other):
        if other is None:
            return False

        if self.__class__ != other.__class__:
            return False

        return self.lineno == other.lineno

    def __hash__(self):
        return self.lineno


class Located(object):
    location = None

    def _error(self, msg, *args):
        '''Shortcut for errors.'''
        location = self.location

        if not location or not location.lineno:
            return msg % args
        s = 'Line %s, %s' % (location.lineno, (msg % args))
        logging.debug(s)
        return s


class Validatable(object):
    is_valid = None
    validated = False
    validation_errors = ()

    def validate(self):
        '''Validate this object and return a list of errors.'''
        if self.validated:
            return self.validation_errors

        errors = self._validate()
        self.is_valid = not errors
        self.validated = True
        self.validation_errors = tuple(errors)

        return self.validation_errors

    def _validate(self):
        '''Override this method in a subclass.'''
        return []



# encoding: utf-8
import logging

from pdefc.lang import references
from pdefc.lang.common import Located, Validatable
from pdefc.lang.types import TypeEnum, Definition


class Message(Definition):
    '''User-defined message.'''
    def __init__(self, name, base=None, discriminator_value=None, declared_fields=None,
                 is_exception=False, doc=None, location=None):
        super(Message, self).__init__(TypeEnum.MESSAGE, name, doc=doc, location=location)

        self.base = base
        self.discriminator_value = discriminator_value
        self._discriminator = None  # Field.

        self.subtypes = []
        self.declared_fields = []
        self.is_exception = is_exception

        if declared_fields:
            for field in declared_fields:
                self.add_field(field)

    @property
    def base(self):
        return self._base.dereference()

    @base.setter
    def base(self, value):
        self._base = references.reference(value)

    @property
    def discriminator_value(self):
        return self._discriminator_value.dereference()

    @discriminator_value.setter
    def discriminator_value(self, value):
        self._discriminator_value = references.reference(value)

    @property
    def discriminator(self):
        '''Return this message discriminator field, base discriminator field, or None.'''
        if self._discriminator:
            return self._discriminator

        return self.base.discriminator if self.base else None

    @property
    def is_polymorphic(self):
        return bool(self.discriminator)

    @property
    def fields(self):
        if not self.base:
            return self.declared_fields

        return self.base.fields + self.declared_fields

    @property
    def inherited_fields(self):
        if not self.base:
            return []

        return self.base.fields

    @property
    def referenced_types(self):
        result = []
        result += self._base.referenced_types
        result += self._discriminator_value.referenced_types
        result += self.subtypes

        for field in self.fields:
            result += field._type.referenced_types

        return [t for t in result if t is not self]

    def add_field(self, field):
        '''Add a new field to this message and return the field.'''
        self.declared_fields.append(field)

        if field.is_discriminator:
            self._discriminator = field

        logging.debug('%s: added a field "%s"', self, field.name)
        return field

    def create_field(self, name, type0, is_discriminator=False):
        '''Create a new field, add it to this message and return the field.'''
        field = Field(name, type0, is_discriminator=is_discriminator)
        return self.add_field(field)

    def _add_subtype(self, subtype):
        '''Add a new subtype to this message.'''
        if not isinstance(subtype, Message):
            raise ValueError('Must be a message instance, %r' % subtype)

        if subtype is self:
            return

        self.subtypes.append(subtype)
        if self.base:
            self.base._add_subtype(subtype)

    def link(self, module):
        errors = super(Message, self).link(module)
        errors += self._base.link(self.lookup)
        errors += self._discriminator_value.link(self.lookup)

        for field in self.declared_fields:
            errors += field.link(self)

        return errors

    def build(self):
        logging.debug('Building %s', self)

        # Add this message to base subtypes.
        if self._discriminator_value and self.base:
            self.base._add_subtype(self)

        return []

    def _validate(self):
        logging.debug('Validation %s', self)

        errors = self._validate_base()
        if errors:
            # Cannot validate this message with an invalid base.
            return errors

        errors = []
        errors += self._validate_discriminator_value()
        errors += self._validate_subtypes()
        errors += self._validate_fields()
        return errors

    @property
    def _has_circular_inheritance(self):
        base = self.base
        while base:
            if base is self:
                return True

            if base.is_message:
                base = base.base
            else:
                # Cannot proceed, the base is not a message.
                break

        return False

    def _validate_base(self):
        if not self.base:
            return []

        # Prevent circular inheritance before any validation.
        if self._has_circular_inheritance:
            # Cannot proceed when circular inheritance.
            return [self._error('%s: circular inheritance', self)]

        # Validate the base, ignore the errors, they are collected
        # when the base is validated in a module.
        base = self.base
        base.validate()

        if not base.is_valid:
            return [self._error('%s: invalid base %s', self, base)]

        if not base.is_message:
            return [self._error('%s: base must be a message but %s', self, base)]

        errors = []
        if self.is_exception != base.is_exception:
            errors.append(self._error('%s: wrong base type (message/exception), %s', self, base))

        if not self._is_defined_after(base):
            errors.append(self._error('%s: must be declared after its base %s', self, base))

        # Base.module can be None in tests.
        if base.module and base.module._depends_on(self.module):
            path = base.module._get_import_path(base.module)
            errors.append(self._error('%s: cannot inherit %s, '
                                      'it is in a dependent module "%s", import path: %s',
                                      self, base, base.module, path))

        return errors

    def _validate_discriminator_value(self):
        base = self.base
        value = self.discriminator_value

        if not base:
            if value:
                return [self._error('%s: discriminator value is present, but no base', self)]

            # No discriminator value, nothing to validate.
            return []

        # The base is present.
        if not value:
            if base.is_polymorphic:
                return [self._error('%s: discriminator value required, base is polymorphic', self)]

            # Non-polymorphic inheritance is correct.
            return []

        # The value is present.
        if not base.is_polymorphic:
            # A present discriminator value requires a polymorphic base.
            return [self._error('%s: cannot set a discriminator value, the base '
                                'is not polymorphic (does not have a discriminator)', self)]

        # Both the value and the base are present, and the base is polymorphic.
        # The base has been validated before, its discriminator field must valid.
        if value not in base.discriminator.type:
            # The discriminator value is not a base discriminator enum value.
            return [self._error('%s: discriminator value does not match '
                                'the base discriminator type', self)]

        enum = value.enum
        if not self._is_defined_after(enum):
            return [self._error('%s: must be declared after discriminator type %s', self, enum)]

        # enum.module can be None in tests.
        if enum.module and enum.module._depends_on(self.module):
            path = enum.module._get_import_path(self.module)
            return [self._error('%s: cannot use %s as a discriminator, '
                                'it is in a dependent module "%s", import path: %s',
                                self, enum, enum.module, path)]

        return []

    def _validate_subtypes(self):
        if not self.subtypes:
            return []

        # Prevent duplicate discriminator values in subtypes.
        errors = []
        values = set()
        for subtype in self.subtypes:
            value = subtype.discriminator_value
            if value in values:
                errors.append(self._error('%s: duplicate subtype with a discriminator value "%s"',
                                          self, value.name))
            values.add(value)

        return errors

    def _validate_fields(self):
        errors = []

        # Prevent duplicate field names.
        names = set()
        for field in self.fields:
            if field.name in names:
                errors.append(self._error('%s: duplicate field "%s"', self, field.name))
            names.add(field.name)

        # Prevent multiple discriminator fields.
        discriminator = None
        for field in self.fields:
            if not field.is_discriminator:
                continue

            if discriminator:
                errors.append(self._error('%s: multiple discriminator fields', self))
                break  # One multiple discriminator error is enough.

            discriminator = field

        for field in self.fields:
            errors += field.validate()

        return errors


class Field(Located, Validatable):
    '''Message field.'''
    def __init__(self, name, type0, is_discriminator=False, location=None):
        self.name = name
        self._type = references.reference(type0)
        self.is_discriminator = is_discriminator
        self.location = location

        self.message = None

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<%s %s at %s>' % (self.__class__.__name__, self.name, hex(id(self)))

    @property
    def type(self):
        return self._type.dereference()

    def link(self, message):
        logging.debug('Linking %s', self)

        if self.message:
            raise ValueError('Field is already linked, %s' % self)
        self.message = message

        return self._type.link(message.lookup)

    def _validate(self):
        logging.debug('Validating %s', self)
        errors = []

        if not self.type.is_data_type:
            errors.append(self._error('%s: field must be a data type', self))

        if self.is_discriminator and not self.type.is_enum:
            errors.append(self._error('%s: discriminator field must be an enum', self))

        # Validate the reference (it can be a collection).
        errors += self._type.validate()
        return errors



# encoding: utf-8
from pdefc.lang.types import Type, TypeEnum


class List(Type):
    '''List collection.'''
    def __init__(self, element, location=None):
        super(List, self).__init__(TypeEnum.LIST, location=location)
        self.element = element

        if not isinstance(element, Type):
            raise TypeError('Element must be a definitions.Type instance, %r' % element)

    def __repr__(self):
        return '<List %s>' % self.element

    def __str__(self):
        return 'list<%s>' % self.element

    def _validate(self):
        errors = []

        if not self.element.is_data_type:
            errors.append(self._error('List element must be a data type'))

        return errors


class Set(Type):
    '''Set collection.'''
    def __init__(self, element, location=None):
        super(Set, self).__init__(TypeEnum.SET, location=location)
        self.element = element

        if not isinstance(element, Type):
            raise TypeError('Element must be a definitions.Type instance, %r' % element)

    def __repr__(self):
        return '<Set %s>' % self.element

    def __str__(self):
        return 'set<%s>' % self.element

    def _validate(self):
        errors = []

        if not self.element.is_data_type:
            errors.append(self._error('Set element must be a data type'))

        return errors


class Map(Type):
    '''Map collection.'''
    def __init__(self, key, value, location=None):
        super(Map, self).__init__(TypeEnum.MAP, location=location)
        self.key = key
        self.value = value

        if not isinstance(key, Type):
            raise TypeError('Key must be a definitions.Type instance, %r' % key)

        if not isinstance(value, Type):
            raise TypeError('Value must be a definitions.Type instance, %r' % value)

    def __repr__(self):
        return '<Map %s, %s>' % (self.key, self.value)

    def __str__(self):
        return 'map<%s, %s>' % (self.key, self.value)

    def _validate(self):
        errors = []

        if not self.key.is_primitive:
            errors.append(self._error('Map key must be a primitive'))

        if not self.value.is_data_type:
            errors.append(self._error('Map value must be a data type'))

        return errors



# encoding: utf-8
from collections import defaultdict
import logging
import re
import yaml


class Package(object):
    '''Protocol definition.'''
    name_pattern = re.compile(r'^[a-zA-Z]{1}[a-zA-Z0-9_]*$')

    def __init__(self, name, info=None, modules=None, dependencies=None):
        self.name = name
        self.info = info
        self.modules = []
        self.dependencies = []

        if modules:
            for module in modules:
                self.add_module(module)

        if dependencies:
            for dep in dependencies:
                self.add_dependency(dep)

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<Package %r at %s>' % (self.name, hex(id(self)))

    def add_dependency(self, package):
        '''Adds all modules from another package to this package dependencies.'''
        if package is self:
            raise ValueError('Cannot include itself %r' % self)

        self.dependencies.append(package)
        logging.debug('%s: added a dependency "%s"', self, package.name)

    def get_dependency(self, name):
        '''Return a package by a name.'''
        for dep in self.dependencies:
            if dep.name == name:
                return dep

    def add_module(self, module):
        '''Add a module to this package.'''
        self.modules.append(module)
        logging.debug('%s: added a module "%s"', self, module.name)

    def get_module(self, relative_name):
        '''Find a module in this package by a name.'''
        for module in self.modules:
            if module.relative_name == relative_name:
                return module

    def lookup_module(self, absolute_name):
        '''Find a module in this package or its dependencies by an absolute name.'''
        if '.' not in absolute_name:
            absolute_name = '%s.%s' % (absolute_name, absolute_name)

        pname, mname = absolute_name.split('.', 1)
        if pname == self.name:
            return self.get_module(mname)

        package = self.get_dependency(pname)
        if package:
            return package.get_module(mname)

        return None

    def compile(self, allow_duplicate_definitions=False):
        '''Compile this package and return a list of errors.'''
        logging.debug('Compiling the package')

        errors = self._link()
        if errors:
            return errors

        errors = self._build()
        if errors:
            return errors

        errors = self._validate(allow_duplicate_definitions)
        if errors:
            return errors

        return []

    def _link(self):
        '''Link this package and return a list of errors.'''
        logging.debug('Linking the package')

        errors = []

        # Prevent duplicate module names.
        names = set()
        for module in self.modules:
            if module.name in names:
                errors.append(self._error('Duplicate module %r', module.name))
            names.add(module.name)

        if errors:
            return errors

        # Link modules.
        for module in self.modules:
            errors += module.link(self)

        return errors

    def _build(self):
        '''Build this package and return a list of errors.'''
        logging.debug('Building the package')

        errors = []
        for module in self.modules:
            errors += module.build()
        return errors

    def _validate(self, allow_duplicate_definitions=False):
        '''Validate this package and return a list of errors.'''
        logging.debug('Validating the package')

        errors = []
        errors += self._validate_name()
        for module in self.modules:
            errors += module.validate()

        if not allow_duplicate_definitions:
            errors += self._validate_no_duplicate_definitions()

        return errors

    def _validate_name(self):
        if not self.name:
            return [self._error('Package name required, add it to the YAML file')]
        if self.name_pattern.match(self.name):
            return []
        return [self._error('Wrong package name "%s". A name must contain only latin letters, '
                            'digits and underscores, and must start with a letter, for example, '
                            '"mycompany_project_api"', self.name)]

    def _validate_no_duplicate_definitions(self):
        errors = []
        names_to_defs = defaultdict(list)

        for module in self.modules:
            for def0 in module.definitions:
                names_to_defs[def0.name].append(def0)

        is_first = True
        for name, defs in names_to_defs.items():
            if len(defs) == 1:
                continue

            if is_first:
                msg = self._error('Duplicate definitions in a package. They are forbidden '
                                  'in languages without namespaces (such as C and Objective C) '
                                  'and will require manual name mapping during code generation. '
                                  'Please, consider renaming them, or explicitly allow them with '
                                  'the --allow-duplicate-definitions flag.')
                errors.append(msg)

            is_first = False
            for def0 in defs:
                path = def0.module.path if def0.module else None
                errors.append(self._error('  %s: %s', def0.name, path))

        return errors

    def _error(self, msg, *args):
        record = msg % args if args else msg
        logging.error(record)
        return record


class PackageInfo(object):
    @classmethod
    def from_yaml(cls, s):
        d = yaml.load(s)
        return PackageInfo.from_dict(d)

    @classmethod
    def from_dict(cls, d):
        p = d.get('package', {})
        name = p.get('name')
        url = p.get('url')
        description = p.get('description')
        modules = p.get('modules')
        dependencies = p.get('dependencies')

        return PackageInfo(name, url=url, description=description, modules=modules,
                           dependencies=dependencies)

    def __init__(self, name, url=None, description=None, modules=None, dependencies=None):
        self.name = name or ''
        self.url = url
        self.description = description or ''
        self.modules = list(modules) if modules else []
        self.dependencies = list(dependencies) if dependencies else []

    def to_dict(self):
        return {'package': {
            'name': self.name,
            'url': self.url,
            'description': self.description,
            'modules': list(self.modules),
            'dependencies': list(self.dependencies)
        }}

    def to_yaml(self):
        return yaml.dump(self.to_dict())



# encoding: utf-8
import os.path
try:
    from setuptools import setup
except ImportError:
    import ez_setup
    ez_setup.use_setuptools()
    from setuptools import setup


# Load the version.
__version__ = None
with open(os.path.join('src', 'pdefc', 'version.py')) as f:
    exec(f.read())


setup(
    name='pdef-compiler',
    version=__version__,
    license='Apache License 2.0',
    url='http://github.com/pdef/pdef',
    description='Pdef compiler',
    long_description=open('README.md', 'r').read(),

    author='Ivan Korobkov',
    author_email='ivan.korobkov@gmail.com',

    packages=['pdefc',
              'pdefc.lang'],
    package_dir={'': 'src'},
    package_data={
        '': ['*.jinja2']
    },

    install_requires=[
        'argparse',
        'jinja2>=2.7',
        'ply>=3.4',
        'PyYAML'
    ],

    tests_require=[
        'mock'
    ],

    entry_points={
        'console_scripts': ['pdefc = pdefc:main'],
        'pdefc.generators': []
    },

    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Software Development :: Compilers'
    ]
)



# encoding: utf-8
import argparse
import logging
import sys

import pdefc
from pdefc.exc import CompilerException


def main(argv=None):
    '''Run the compiler command-line interface.'''
    cli = Cli()
    cli.run(argv)


class Cli(object):
    '''Pdef command-line interface.'''
    def _create_compiler(self, paths=None, allow_duplicate_definitions=False):
        return pdefc.create_compiler(paths, allow_duplicate_definitions=allow_duplicate_definitions)

    def _find_generators(self):
        return pdefc.find_generators()

    def run(self, argv=None):
        # Configure logging before the commands, because the latter
        # requires a functional logger.
        self._logging(argv)

        try:
            args = self._parse(argv)

            # The command_func is set in each subparser.
            func = args.command_func
            if not func:
                raise ValueError('No command_func in %s' % args)

            return func(args)

        except pdefc.CompilerException as e:
            # It's an expected exception.
            # Get rid of the traceback.
            logging.error('%s', e)
            sys.exit(1)

    def _parse(self, argv):
        parser = self._create_parser()
        args = parser.parse_args(argv)

        logging.debug('Arguments: %s', args)
        return args

    def _create_parser(self):
        parser = argparse.ArgumentParser(description='Pdef compiler, see http://github.com/pdef')
        subparsers = parser.add_subparsers(dest='command', title='commands',
            description='To show a command help execute "pdefc {command} -h"')

        self._logging_args(parser)
        self._check_args(subparsers)
        self._generate_args(subparsers)
        return parser

    # Logging.

    def _logging(self, argv=None):
        argv = argv or sys.argv

        level = logging.WARN
        if ('-v' in argv) or ('--verbose' in argv):
            level = logging.INFO
        elif '--debug' in argv:
            level = logging.DEBUG

        logging.basicConfig(level=level, format='%(message)s')

    def _logging_args(self, parser):
        parser.add_argument('-v', '--verbose', action='store_true', help='verbose output')
        parser.add_argument('--debug', action='store_true', help='debug output')

    # Check.

    def _check(self, args):
        package = args.package
        paths = args.paths
        allow_duplicate_definitions = args.allow_duplicate_definitions

        compiler = self._create_compiler(paths, allow_duplicate_definitions)
        return compiler.check(package)

    def _check_args(self, subparsers):
        parser = subparsers.add_parser('check', help='check a package')
        parser.add_argument('package', help='path to a package yaml file')
        parser.add_argument('--include', dest='paths', action='append', default=[],
                            help='paths to package dependencies')
        parser.add_argument('--allow-duplicate-definitions', dest='allow_duplicate_definitions',
                            action='store_true', default=False,
                            help='allow duplicate definition names in a package')
        parser.set_defaults(command_func=self._check)

    # Generate.

    def _generate(self, args):
        package = args.package
        generator = args.generator
        paths = args.paths
        out = args.out
        namespace = self._parse_namespace(args.namespace)
        allow_duplicate_definitions = args.allow_duplicate_definitions

        compiler = self._create_compiler(paths, allow_duplicate_definitions)
        return compiler.generate(package, generator, out=out, namespace=namespace)

    def _generate_args(self, subparsers):
        generators = self._find_generators()
        generator_names = list(generators.keys())
        description = self._generators_description()

        parser = subparsers.add_parser('generate', description=description,
                                       formatter_class=argparse.RawDescriptionHelpFormatter,
                                       help='generate source code from a package')
        parser.add_argument('package', help='path to a package yaml file')
        parser.add_argument('--generator', choices=generator_names, required=True)
        parser.add_argument('--out', dest='out', required=True,
                            help='directory for generated files')
        parser.add_argument('--ns', dest='namespace', action='append', default=[],
                            help='adds a namespace which maps pdef modules '
                                 'to generated modules, i.e. "pdef.module:io.pdef.java"')
        parser.add_argument('--include', dest='paths', action='append', default=[],
                            help='paths to package dependencies')
        parser.add_argument('--allow-duplicate-definitions', dest='allow_duplicate_definitions',
                            action='store_true', default=False,
                            help='allow duplicate definition names in a package')

        parser.set_defaults(command_func=self._generate)

    def _generators_description(self):
        '''Return a generators description string.'''
        generators = dict(self._find_generators())

        description  = ['available generators:']

        for name, generator in generators.items():
            doc = generator.__doc__ or ''

            first = True
            for line in doc.splitlines():
                if first:
                    line = line.strip() or 'no description'
                    description.append('  - %s: %s' % (name, line))
                    first = False
                elif line.strip():
                    description.append(line)

            if first:
                description.append('  - %s: no description' % name)

        description.append('\nmore generators: \n  see https://github.com/pdef/pdef')
        return '\n'.join(description)

    def _parse_namespace(self, seq):
        if not seq:
            return {}

        result = {}
        error = False
        for item in seq:
            parts = item.split(':')
            if len(parts) != 2:
                logging.error('Wrong namespace "%s", the namespace must be specified as '
                              '"pdef.module:lang.module"', item)
                error = True
                continue

            pmodule, lmodule = parts
            result[pmodule] = lmodule

        if error:
            raise CompilerException('Wrong arguments')

        logging.debug('Namespace %s', result)
        return result



# encoding: utf-8
import logging

from pdefc.lang import references
from pdefc.lang.common import Located, Validatable
from pdefc.lang.types import TypeEnum, NativeType, Definition


class Interface(Definition):
    '''User-defined interface.'''
    def __init__(self, name, exc=None, methods=None, doc=None, location=None):
        super(Interface, self).__init__(TypeEnum.INTERFACE, name, doc=doc, location=location)

        self.exc = exc
        self.methods = []

        if methods:
            for method in methods:
                self.add_method(method)

    @property
    def exc(self):
        return self._exc.dereference()

    @exc.setter
    def exc(self, value):
        self._exc = references.reference(value)

    @property
    def declared_methods(self):
        return self.methods

    @property
    def referenced_types(self):
        result = []
        result += self._exc.referenced_types

        for method in self.methods:
            result += method._result.referenced_types
            for arg in method.args:
                result += arg._type.referenced_types

        return [t for t in result if t is not self]

    def add_method(self, method):
        '''Add a method to this interface.'''
        self.methods.append(method)
        logging.debug('%s: added a method "%s"', self, method.name)

    def create_method(self, name, result=NativeType.VOID, arg_tuples=None, is_post=False):
        '''Add a new method to this interface and return the method.'''
        method = Method(name, result=result, is_post=is_post)
        if arg_tuples:
            for arg_tuple in arg_tuples:
                method.create_arg(*arg_tuple)

        self.add_method(method)
        return method

    def link(self, module):
        '''Link the base, the exception and the methods.'''
        errors = super(Interface, self).link(module)
        errors += self._exc.link(module.lookup)

        for method in self.methods:
            errors += method.link(self)

        return errors

    def _validate(self):
        logging.debug('Validating %s', self)
        errors = []
        errors += self._validate_methods()
        errors += self._validate_exc()
        return errors

    def _validate_exc(self):
        if not self.exc:
            return []

        if not self.exc.is_exception:
            return [self._error('%s: exc must be an exception, got %s', self, self.exc)]

        return self._exc.validate()

    def _validate_methods(self):
        errors = []

        # Prevent duplicate methods.
        names = set()
        for method in self.methods:
            if method.name in names:
                errors.append(self._error('%s: duplicate method "%s"', self, method.name))
            names.add(method.name)

        # Validate methods.
        for method in self.methods:
            errors += method.validate()

        return errors


class Method(Located, Validatable):
    '''Interface method.'''
    def __init__(self, name, result=NativeType.VOID, args=None, is_post=False, doc=None,
                 location=None):
        self.name = name
        self.args = []
        self.result = result
        self.interface = None
        self.is_post = is_post

        self.doc = doc
        self.location = location

        if args:
            for arg in args:
                self.add_arg(arg)

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<%s %s at %s>' % (self.__class__.__name__, self.name, hex(id(self)))

    @property
    def result(self):
        return self._result.dereference()

    @result.setter
    def result(self, value):
        self._result = references.reference(value)

    @property
    def is_terminal(self):
        return self.result and self.result.type != TypeEnum.INTERFACE

    def lookup(self, name):
        return self.interface.lookup(name) if self.interface else None

    def add_arg(self, arg):
        '''Append an argument to this method.'''
        if arg.method:
            raise ValueError('Argument is already in a method, %s' % arg)

        self.args.append(arg)
        logging.debug('%s: added an arg "%s"', self, arg.name)

    def create_arg(self, name, definition, is_query=False, is_post=False):
        '''Create a new arg and add it to this method.'''
        arg = MethodArg(name, definition, is_query=is_query, is_post=is_post)
        self.add_arg(arg)
        return arg

    def link(self, interface):
        logging.debug('Linking %s', self)

        if self.interface:
            raise ValueError('Method is already linked, method=%s', self)
        self.interface = interface

        errors = []
        errors += self._result.link(self.lookup)
        for arg in self.args:
            errors += arg.link(self)

        return errors

    def _validate(self):
        logging.debug('Validating %s', self)
        errors = []

        # The method must have a result.
        if not self.result:
            errors.append(self._error('%s: method result required', self))
        else:
            errors += self._result.validate()

        # @post methods must be terminal.
        if self.is_post and not self.is_terminal:
            errors.append(self._error('%s: @post method must be terminal (return a data type '
                                      'or void)', self))

        # Prevent duplicate arguments.
        names = set()
        for arg in self.args:
            if arg.name in names:
                errors.append(self._error('%s: duplicate argument "%s"', self, arg.name))
            names.add(arg.name)

        # Validate post/query arguments.
        is_post = self.is_post
        is_nonpost_terminal = not is_post and self.is_terminal
        for arg in self.args:
            if arg.is_post and not is_post:
                errors.append(self._error(
                    '%s: @post arguments can be declared only in @post methods', self))
                break

            elif arg.is_query and not is_nonpost_terminal:
                errors.append(
                    self._error('%s: @query arguments can be declared only in terminal non-post '
                                'methods', self))
                break

        # Validate arguments.
        for arg in self.args:
            errors += arg.validate()

        return errors


class MethodArg(Located, Validatable):
    '''Single method argument.'''
    def __init__(self, name, type0, is_query=False, is_post=False):
        self.name = name
        self.type = type0
        self.method = None

        self.is_query = is_query
        self.is_post = is_post

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<%s %s at %s>' % (self.__class__.__name__, self.name, hex(id(self)))

    @property
    def type(self):
        return self._type.dereference()

    @type.setter
    def type(self, value):
        self._type = references.reference(value)

    @property
    def fullname(self):
        return '%s.%s' % (self.method, self.name)

    def link(self, method):
        if self.method:
            raise ValueError('Method argument is already linked, %s' % self)
        self.method = method

        return self._type.link(method.lookup)

    def _validate(self):
        if not self.type:
            return [self._error('%s: argument type required', self)]

        if not self.type.is_data_type:
            return [self._error('%s: argument must be a data type', self)]

        errors = []
        if self.is_post and self.is_query:
            errors = [self._error('%s: argument cannot be both @query and @post', self)]

        errors += self._type.validate()
        return errors
